@@ -1,46 +1,103 @@
 // SPDX-License-Identifier: GPL-2.0-only
 #define _POSIX_C_SOURCE 200809L
+#include "xwayland.h"
 #include <assert.h>
 #include <stdlib.h>
+#include <wlr/types/wlr_compositor.h>
+#include <wlr/types/wlr_output_layout.h>
+#include <wlr/types/wlr_scene.h>
+#include <wlr/types/wlr_seat.h>
+#include <wlr/types/wlr_xcursor_manager.h>
 #include <wlr/xwayland.h>
+#include "buffer.h"
 #include "common/array.h"
 #include "common/macros.h"
 #include "common/mem.h"
 #include "config/rcxml.h"
 #include "config/session.h"
-#include "foreign-toplevel.h"
+#include "foreign-toplevel/foreign.h"
 #include "labwc.h"
 #include "node.h"
-#include "ssd.h"
+#include "output.h"
 #include "view.h"
 #include "view-impl-common.h"
 #include "window-rules.h"
 #include "workspaces.h"
-#include "xwayland.h"
 
-xcb_atom_t atoms[WINDOW_TYPE_LEN] = {0};
+enum atoms {
+	ATOM_NET_WM_ICON = 0,
 
-static void xwayland_view_unmap(struct view *view, bool client_request);
+	ATOM_COUNT,
+};
 
-static bool
-xwayland_surface_contains_window_type(
-		struct wlr_xwayland_surface *surface, enum window_type window_type)
+static const char * const atom_names[] = {
+	[ATOM_NET_WM_ICON] = "_NET_WM_ICON",
+};
+
+static_assert(ARRAY_SIZE(atom_names) == ATOM_COUNT, "atom names out of sync");
+
+static xcb_atom_t atoms[ATOM_COUNT] = {0};
+
+static void set_surface(struct view *view, struct wlr_surface *surface);
+static void handle_map(struct wl_listener *listener, void *data);
+static void handle_unmap(struct wl_listener *listener, void *data);
+
+static struct xwayland_view *
+xwayland_view_from_view(struct view *view)
 {
-	assert(surface);
-	for (size_t i = 0; i < surface->window_type_len; i++) {
-		if (surface->window_type[i] == atoms[window_type]) {
-			return true;
-		}
-	}
-	return false;
+	assert(view->type == LAB_XWAYLAND_VIEW);
+	return (struct xwayland_view *)view;
+}
+
+static struct wlr_xwayland_surface *
+xwayland_surface_from_view(struct view *view)
+{
+	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
+	assert(xwayland_view->xwayland_surface);
+	return xwayland_view->xwayland_surface;
 }
 
 static bool
-xwayland_view_contains_window_type(struct view *view, int32_t window_type)
+xwayland_view_contains_window_type(struct view *view,
+		enum lab_window_type window_type)
 {
+	/* Compile-time check that the enum types match */
+	static_assert(LAB_WINDOW_TYPE_DESKTOP ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DESKTOP
+		&& LAB_WINDOW_TYPE_DOCK ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DOCK
+		&& LAB_WINDOW_TYPE_TOOLBAR ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_TOOLBAR
+		&& LAB_WINDOW_TYPE_MENU ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_MENU
+		&& LAB_WINDOW_TYPE_UTILITY ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_UTILITY
+		&& LAB_WINDOW_TYPE_SPLASH ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_SPLASH
+		&& LAB_WINDOW_TYPE_DIALOG ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DIALOG
+		&& LAB_WINDOW_TYPE_DROPDOWN_MENU ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DROPDOWN_MENU
+		&& LAB_WINDOW_TYPE_POPUP_MENU ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_POPUP_MENU
+		&& LAB_WINDOW_TYPE_TOOLTIP ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_TOOLTIP
+		&& LAB_WINDOW_TYPE_NOTIFICATION ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NOTIFICATION
+		&& LAB_WINDOW_TYPE_COMBO ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_COMBO
+		&& LAB_WINDOW_TYPE_DND ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DND
+		&& LAB_WINDOW_TYPE_NORMAL ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NORMAL
+		&& LAB_WINDOW_TYPE_LEN ==
+			(int)WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NORMAL + 1,
+		"lab_window_type does not match wlr_xwayland_net_wm_window_type");
+
 	assert(view);
 	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(view);
-	return xwayland_surface_contains_window_type(surface, window_type);
+	return wlr_xwayland_surface_has_window_type(surface,
+		(enum wlr_xwayland_net_wm_window_type)window_type);
 }
 
 static struct view_size_hints
@@ -66,7 +123,7 @@ xwayland_view_wants_focus(struct view *view)
 	struct wlr_xwayland_surface *xsurface =
 		xwayland_surface_from_view(view);
 
-	switch (wlr_xwayland_icccm_input_model(xsurface)) {
+	switch (wlr_xwayland_surface_icccm_input_model(xsurface)) {
 	/*
 	 * Abbreviated from ICCCM section 4.1.7 (Input Focus):
 	 *
@@ -94,24 +151,19 @@ xwayland_view_wants_focus(struct view *view)
 	 * Globally Active client windows may receive a WM_TAKE_FOCUS
 	 * message from the window manager. If they want the focus, they
 	 * should respond with a SetInputFocus request.
-	 *
-	 * [Currently, labwc does not fully support clients voluntarily
-	 * taking focus via the WM_TAKE_FOCUS + SetInputFocus mechanism.
-	 * Instead, we try to guess whether the window wants focus based
-	 * on some heuristics -- see below.]
 	 */
 	case WLR_ICCCM_INPUT_MODEL_GLOBAL:
 		/*
-		 * Assume that NORMAL and DIALOG windows always want
-		 * focus. These window types should show up in the
+		 * Assume that NORMAL and DIALOG windows are likely to
+		 * want focus. These window types should show up in the
 		 * Alt-Tab switcher and be automatically focused when
 		 * they become topmost.
 		 */
-		return (xwayland_surface_contains_window_type(xsurface,
-				NET_WM_WINDOW_TYPE_NORMAL)
-			|| xwayland_surface_contains_window_type(xsurface,
-				NET_WM_WINDOW_TYPE_DIALOG)) ?
-			VIEW_WANTS_FOCUS_ALWAYS : VIEW_WANTS_FOCUS_OFFER;
+		return (wlr_xwayland_surface_has_window_type(xsurface,
+				WLR_XWAYLAND_NET_WM_WINDOW_TYPE_NORMAL)
+			|| wlr_xwayland_surface_has_window_type(xsurface,
+				WLR_XWAYLAND_NET_WM_WINDOW_TYPE_DIALOG)) ?
+			VIEW_WANTS_FOCUS_LIKELY : VIEW_WANTS_FOCUS_UNLIKELY;
 
 	/*
 	 * No Input - The client never expects keyboard input.
@@ -135,6 +187,19 @@ xwayland_view_has_strut_partial(struct view *view)
 	return (bool)xsurface->strut_partial;
 }
 
+static void
+xwayland_view_offer_focus(struct view *view)
+{
+	wlr_xwayland_surface_offer_focus(xwayland_surface_from_view(view));
+}
+
+static struct view *
+xwayland_view_get_parent(struct view *view)
+{
+	struct wlr_xwayland_surface *xsurface = xwayland_surface_from_view(view);
+	return xsurface->parent ? (struct view *)xsurface->parent->data : NULL;
+}
+
 static struct wlr_xwayland_surface *
 top_parent_of(struct view *view)
 {
@@ -145,21 +210,6 @@ top_parent_of(struct view *view)
 	return s;
 }
 
-static struct xwayland_view *
-xwayland_view_from_view(struct view *view)
-{
-	assert(view->type == LAB_XWAYLAND_VIEW);
-	return (struct xwayland_view *)view;
-}
-
-struct wlr_xwayland_surface *
-xwayland_surface_from_view(struct view *view)
-{
-	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
-	assert(xwayland_view->xwayland_surface);
-	return xwayland_view->xwayland_surface;
-}
-
 static void
 ensure_initial_geometry_and_output(struct view *view)
 {
@@ -178,7 +228,7 @@ ensure_initial_geometry_and_output(struct view *view)
 			view->pending = view->current;
 		}
 	}
-	if (!view->output) {
+	if (!output_is_usable(view->output)) {
 		/*
 		 * Just use the cursor output since we don't know yet
 		 * whether the surface position is meaningful.
@@ -224,6 +274,7 @@ handle_commit(struct wl_listener *listener, void *data)
 	 */
 	if (current->width != state->width || current->height != state->height) {
 		view_impl_apply_geometry(view, state->width, state->height);
+		view_moved(view);
 	}
 }
 
@@ -237,11 +288,11 @@ handle_request_move(struct wl_listener *listener, void *data)
 	 * the provided serial against a list of button press serials sent to
 	 * this client, to prevent the client from requesting this whenever they
 	 * want.
+	 *
+	 * Note: interactive_begin() checks that view == server->grabbed_view.
 	 */
 	struct view *view = wl_container_of(listener, view, request_move);
-	if (view == view->server->seat.pressed.view) {
-		interactive_begin(view, LAB_INPUT_STATE_MOVE, 0);
-	}
+	interactive_begin(view, LAB_INPUT_STATE_MOVE, LAB_EDGE_NONE);
 }
 
 static void
@@ -254,12 +305,12 @@ handle_request_resize(struct wl_listener *listener, void *data)
 	 * the provided serial against a list of button press serials sent to
 	 * this client, to prevent the client from requesting this whenever they
 	 * want.
+	 *
+	 * Note: interactive_begin() checks that view == server->grabbed_view.
 	 */
 	struct wlr_xwayland_resize_event *event = data;
 	struct view *view = wl_container_of(listener, view, request_resize);
-	if (view == view->server->seat.pressed.view) {
-		interactive_begin(view, LAB_INPUT_STATE_RESIZE, event->edges);
-	}
+	interactive_begin(view, LAB_INPUT_STATE_RESIZE, event->edges);
 }
 
 static void
@@ -270,7 +321,7 @@ handle_associate(struct wl_listener *listener, void *data)
 	assert(xwayland_view->xwayland_surface &&
 		xwayland_view->xwayland_surface->surface);
 
-	view_connect_map(&xwayland_view->base,
+	set_surface(&xwayland_view->base,
 		xwayland_view->xwayland_surface->surface);
 }
 
@@ -280,19 +331,7 @@ handle_dissociate(struct wl_listener *listener, void *data)
 	struct xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, dissociate);
 
-	/* https://gitlab.freedesktop.org/wlroots/wlroots/-/merge_requests/4524 */
-	assert(xwayland_view->base.mappable.connected);
-	mappable_disconnect(&xwayland_view->base.mappable);
-}
-
-static void
-handle_surface_destroy(struct wl_listener *listener, void *data)
-{
-	struct view *view = wl_container_of(listener, view, surface_destroy);
-	assert(data && data == view->surface);
-
-	view->surface = NULL;
-	wl_list_remove(&view->surface_destroy.link);
+	set_surface(&xwayland_view->base, NULL);
 }
 
 static void
@@ -302,15 +341,7 @@ handle_destroy(struct wl_listener *listener, void *data)
 	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
 	assert(xwayland_view->xwayland_surface->data == view);
 
-	if (view->surface) {
-		/*
-		 * We got the destroy signal from
-		 * wlr_xwayland_surface before the
-		 * destroy signal from wlr_surface.
-		 */
-		wl_list_remove(&view->surface_destroy.link);
-	}
-	view->surface = NULL;
+	set_surface(view, NULL);
 
 	/*
 	 * Break view <-> xsurface association.  Note that the xsurface
@@ -331,6 +362,7 @@ handle_destroy(struct wl_listener *listener, void *data)
 	wl_list_remove(&xwayland_view->set_override_redirect.link);
 	wl_list_remove(&xwayland_view->set_strut_partial.link);
 	wl_list_remove(&xwayland_view->set_window_type.link);
+	wl_list_remove(&xwayland_view->focus_in.link);
 	wl_list_remove(&xwayland_view->map_request.link);
 
 	view_destroy(view);
@@ -417,19 +449,28 @@ static void
 handle_request_maximize(struct wl_listener *listener, void *data)
 {
 	struct view *view = wl_container_of(listener, view, request_maximize);
+	struct wlr_xwayland_surface *surf = xwayland_surface_from_view(view);
 	if (!view->mapped) {
 		ensure_initial_geometry_and_output(view);
 		/*
 		 * Set decorations early to avoid changing geometry
 		 * after maximize (reduces visual glitches).
 		 */
-		if (want_deco(xwayland_surface_from_view(view))) {
+		if (want_deco(surf)) {
 			view_set_ssd_mode(view, LAB_SSD_MODE_FULL);
 		} else {
 			view_set_ssd_mode(view, LAB_SSD_MODE_NONE);
 		}
 	}
-	view_toggle_maximize(view, VIEW_AXIS_BOTH);
+
+	enum view_axis maximize = VIEW_AXIS_NONE;
+	if (surf->maximized_vert) {
+		maximize |= VIEW_AXIS_VERTICAL;
+	}
+	if (surf->maximized_horz) {
+		maximize |= VIEW_AXIS_HORIZONTAL;
+	}
+	view_maximize(view, maximize);
 }
 
 static void
@@ -447,7 +488,8 @@ static void
 handle_set_title(struct wl_listener *listener, void *data)
 {
 	struct view *view = wl_container_of(listener, view, set_title);
-	view_update_title(view);
+	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
+	view_set_title(view, xwayland_view->xwayland_surface->title);
 }
 
 static void
@@ -456,35 +498,7 @@ handle_set_class(struct wl_listener *listener, void *data)
 	struct xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, set_class);
 	struct view *view = &xwayland_view->base;
-	view_update_app_id(view);
-}
 
-static void
-xwayland_view_close(struct view *view)
-{
-	wlr_xwayland_surface_close(xwayland_surface_from_view(view));
-}
-
-static const char *
-xwayland_view_get_string_prop(struct view *view, const char *prop)
-{
-	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
-	struct wlr_xwayland_surface *xwayland_surface = xwayland_view->xwayland_surface;
-	if (!xwayland_surface) {
-		/*
-		 * This may happen due to a matchOnce rule when
-		 * a view is destroyed while A-Tab is open. See
-		 * https://github.com/labwc/labwc/issues/1082#issuecomment-1716137180
-		 */
-		return "";
-	}
-
-	if (!strcmp(prop, "title")) {
-		return xwayland_surface->title ? xwayland_surface->title : "";
-	}
-	if (!strcmp(prop, "class")) {
-		return xwayland_surface->class ? xwayland_surface->class : "";
-	}
 	/*
 	 * Use the WM_CLASS 'instance' (1st string) for the app_id. Per
 	 * ICCCM, this is usually "the trailing part of the name used to
@@ -494,10 +508,13 @@ xwayland_view_get_string_prop(struct view *view, const char *prop)
 	 * 'instance' except for being capitalized. We want lowercase
 	 * here since we use the app_id for icon lookups.
 	 */
-	if (!strcmp(prop, "app_id")) {
-		return xwayland_surface->instance ? xwayland_surface->instance : "";
-	}
-	return "";
+	view_set_app_id(view, xwayland_view->xwayland_surface->instance);
+}
+
+static void
+xwayland_view_close(struct view *view)
+{
+	wlr_xwayland_surface_close(xwayland_surface_from_view(view));
 }
 
 static void
@@ -531,7 +548,7 @@ handle_set_override_redirect(struct wl_listener *listener, void *data)
 	struct server *server = view->server;
 	bool mapped = xsurface->surface && xsurface->surface->mapped;
 	if (mapped) {
-		xwayland_view_unmap(view, /* client_request */ true);
+		handle_unmap(&view->mappable.unmap, NULL);
 	}
 	handle_destroy(&view->destroy, xsurface);
 	/* view is invalid after this point */
@@ -550,6 +567,94 @@ handle_set_strut_partial(struct wl_listener *listener, void *data)
 	}
 }
 
+static void
+update_icon(struct xwayland_view *xwayland_view)
+{
+	if (!xwayland_view->xwayland_surface) {
+		return;
+	}
+
+	xcb_window_t window_id = xwayland_view->xwayland_surface->window_id;
+
+	xcb_connection_t *xcb_conn = wlr_xwayland_get_xwm_connection(
+		xwayland_view->base.server->xwayland);
+	xcb_get_property_cookie_t cookie = xcb_get_property(xcb_conn, 0,
+		window_id, atoms[ATOM_NET_WM_ICON], XCB_ATOM_CARDINAL, 0, 0x10000);
+	xcb_get_property_reply_t *reply = xcb_get_property_reply(xcb_conn, cookie, NULL);
+	if (!reply) {
+		return;
+	}
+	xcb_ewmh_get_wm_icon_reply_t icon;
+	if (!xcb_ewmh_get_wm_icon_from_reply(&icon, reply)) {
+		wlr_log(WLR_INFO, "Invalid x11 icon");
+		view_set_icon(&xwayland_view->base, NULL, NULL);
+		goto out;
+	}
+
+	xcb_ewmh_wm_icon_iterator_t iter = xcb_ewmh_get_wm_icon_iterator(&icon);
+	struct wl_array buffers;
+	wl_array_init(&buffers);
+	for (; iter.rem; xcb_ewmh_get_wm_icon_next(&iter)) {
+		size_t stride = iter.width * 4;
+		uint32_t *buf = xzalloc(iter.height * stride);
+
+		/* Pre-multiply alpha */
+		for (uint32_t y = 0; y < iter.height; y++) {
+			for (uint32_t x = 0; x < iter.width; x++) {
+				uint32_t i = x + y * iter.width;
+				uint8_t *src_pixel = (uint8_t *)&iter.data[i];
+				uint8_t *dst_pixel = (uint8_t *)&buf[i];
+				dst_pixel[0] = src_pixel[0] * src_pixel[3] / 255;
+				dst_pixel[1] = src_pixel[1] * src_pixel[3] / 255;
+				dst_pixel[2] = src_pixel[2] * src_pixel[3] / 255;
+				dst_pixel[3] = src_pixel[3];
+			}
+		}
+
+		struct lab_data_buffer *buffer = buffer_create_from_data(
+			buf, iter.width, iter.height, stride);
+		array_add(&buffers, buffer);
+	}
+
+	/* view takes ownership of the buffers */
+	view_set_icon(&xwayland_view->base, NULL, &buffers);
+	wl_array_release(&buffers);
+
+out:
+	free(reply);
+}
+
+static void
+handle_focus_in(struct wl_listener *listener, void *data)
+{
+	struct xwayland_view *xwayland_view =
+		wl_container_of(listener, xwayland_view, focus_in);
+	struct view *view = &xwayland_view->base;
+	struct seat *seat = &view->server->seat;
+
+	if (!view->surface) {
+		/*
+		 * It is rare but possible for the focus_in event to be
+		 * received before the map event. This has been seen
+		 * during CLion startup, when focus is initially offered
+		 * to the splash screen but accepted later by the main
+		 * window instead. (In this case, the focus transfer is
+		 * client-initiated but allowed by wlroots because the
+		 * same PID owns both windows.)
+		 *
+		 * Set a flag to record this condition, and update the
+		 * seat focus later when the view is actually mapped.
+		 */
+		wlr_log(WLR_DEBUG, "focus_in received before map");
+		xwayland_view->focused_before_map = true;
+		return;
+	}
+
+	if (view->surface != seat->seat->keyboard_state.focused_surface) {
+		seat_focus_surface(seat, view->surface);
+	}
+}
+
 /*
  * Sets the initial geometry of maximized/fullscreen views before
  * actually mapping them, so that they can do their initial layout and
@@ -599,7 +704,7 @@ handle_map_request(struct wl_listener *listener, void *data)
 	if (xsurface->maximized_vert) {
 		axis |= VIEW_AXIS_VERTICAL;
 	}
-	view_maximize(view, axis, /*store_natural_geometry*/ true);
+	view_maximize(view, axis);
 	/*
 	 * We could also call set_initial_position() here, but it's not
 	 * really necessary until the view is actually mapped (and at
@@ -610,17 +715,15 @@ handle_map_request(struct wl_listener *listener, void *data)
 static void
 check_natural_geometry(struct view *view)
 {
-	int min_width = view_get_min_width();
-
 	/*
 	 * Some applications (example: Thonny) don't set a reasonable
 	 * un-maximized size when started maximized. Try to detect this
 	 * and set a fallback size.
 	 */
 	if (!view_is_floating(view)
-			&& (view->natural_geometry.width < min_width
+			&& (view->natural_geometry.width < LAB_MIN_VIEW_WIDTH
 			|| view->natural_geometry.height < LAB_MIN_VIEW_HEIGHT)) {
-		view_set_fallback_natural_geometry(view);
+		view->natural_geometry = view_get_fallback_natural_geometry(view);
 	}
 }
 
@@ -663,43 +766,36 @@ set_initial_position(struct view *view,
 }
 
 static void
-init_foreign_toplevel(struct view *view)
+set_surface(struct view *view, struct wlr_surface *surface)
 {
-	assert(!view->foreign_toplevel);
-	view->foreign_toplevel = foreign_toplevel_create(view);
-
-	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(view);
-	if (!surface->parent) {
-		return;
+	if (view->surface) {
+		/* Disconnect wlr_surface event listeners */
+		mappable_disconnect(&view->mappable);
+		wl_list_remove(&view->commit.link);
 	}
-	struct view *parent = (struct view *)surface->parent->data;
-	if (!parent || !parent->foreign_toplevel) {
-		return;
+	view->surface = surface;
+	if (surface) {
+		/* Connect wlr_surface event listeners */
+		mappable_connect(&view->mappable, surface,
+			handle_map, handle_unmap);
+		CONNECT_SIGNAL(surface, view, commit);
 	}
-	foreign_toplevel_set_parent(view->foreign_toplevel, parent->foreign_toplevel);
 }
 
 static void
-xwayland_view_map(struct view *view)
+handle_map(struct wl_listener *listener, void *data)
 {
+	struct view *view = wl_container_of(listener, view, mappable.map);
 	struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
 	struct wlr_xwayland_surface *xwayland_surface =
 		xwayland_view->xwayland_surface;
 	assert(xwayland_surface);
+	assert(xwayland_surface->surface);
+	assert(xwayland_surface->surface == view->surface);
 
 	if (view->mapped) {
 		return;
 	}
-	if (!xwayland_surface->surface) {
-		/*
-		 * We may get here if a user minimizes an xwayland dialog at the
-		 * same time as the client requests unmap, which xwayland
-		 * clients sometimes do without actually requesting destroy
-		 * even if they don't intend to use that view/surface anymore
-		 */
-		wlr_log(WLR_DEBUG, "Cannot map view without wlr_surface");
-		return;
-	}
 
 	/*
 	 * The map_request event may not be received when an unmanaged
@@ -710,41 +806,18 @@ xwayland_view_map(struct view *view)
 	handle_map_request(&xwayland_view->map_request, NULL);
 
 	view->mapped = true;
-	wlr_scene_node_set_enabled(&view->scene_tree->node, true);
 
-	if (view->surface != xwayland_surface->surface) {
-		if (view->surface) {
-			wl_list_remove(&view->surface_destroy.link);
-		}
-		view->surface = xwayland_surface->surface;
-
-		/* Required to set the surface to NULL when destroyed by the client */
-		view->surface_destroy.notify = handle_surface_destroy;
-		wl_signal_add(&view->surface->events.destroy, &view->surface_destroy);
-
-		/* Will be free'd automatically once the surface is being destroyed */
-		struct wlr_scene_tree *tree = wlr_scene_subsurface_tree_create(
+	if (!view->content_tree) {
+		view->content_tree = wlr_scene_subsurface_tree_create(
 			view->scene_tree, view->surface);
-		if (!tree) {
+		if (!view->content_tree) {
 			/* TODO: might need further clean up */
 			wl_resource_post_no_memory(view->surface->resource);
 			return;
 		}
-		view->content_tree = tree;
 	}
 
-	/*
-	 * Exclude unfocusable views from wlr-foreign-toplevel. These
-	 * views (notifications, floating toolbars, etc.) should not be
-	 * shown in taskbars/docks/etc.
-	 */
-	if (!view->foreign_toplevel && view_is_focusable(view)) {
-		init_foreign_toplevel(view);
-		/*
-		 * Initial outputs will be synced via
-		 * view->events.new_outputs on view_moved()
-		 */
-	}
+	wlr_scene_node_set_enabled(&view->content_tree->node, !view->shaded);
 
 	if (!view->been_mapped) {
 		check_natural_geometry(view);
@@ -759,52 +832,49 @@ xwayland_view_map(struct view *view)
 		view_moved(view);
 	}
 
-	/* Add commit here, as xwayland map/unmap can change the wlr_surface */
-	wl_signal_add(&xwayland_surface->surface->events.commit, &view->commit);
-	view->commit.notify = handle_commit;
+	/*
+	 * If the view was focused (on the xwayland server side) before
+	 * being mapped, update the seat focus now. Note that this only
+	 * really matters in the case of Globally Active input windows.
+	 * In all other cases, it's redundant since view_impl_map()
+	 * results in the view being focused anyway.
+	 */
+	if (xwayland_view->focused_before_map) {
+		xwayland_view->focused_before_map = false;
+		seat_focus_surface(&view->server->seat, view->surface);
+	}
 
 	view_impl_map(view);
 	view->been_mapped = true;
-
-	/* Update usable area to account for XWayland "struts" (panels) */
-	if (xwayland_surface->strut_partial) {
-		output_update_all_usable_areas(view->server, false);
-	}
 }
 
 static void
-xwayland_view_unmap(struct view *view, bool client_request)
+handle_unmap(struct wl_listener *listener, void *data)
 {
+	struct view *view = wl_container_of(listener, view, mappable.unmap);
 	if (!view->mapped) {
-		goto out;
+		return;
 	}
 	view->mapped = false;
-	wl_list_remove(&view->commit.link);
-	wlr_scene_node_set_enabled(&view->scene_tree->node, false);
 	view_impl_unmap(view);
 
-	/* Update usable area to account for XWayland "struts" (panels) */
-	if (xwayland_surface_from_view(view)->strut_partial) {
-		output_update_all_usable_areas(view->server, false);
-	}
-
 	/*
-	 * If the view was explicitly unmapped by the client (rather
-	 * than just minimized), destroy the foreign toplevel handle so
-	 * the unmapped view doesn't show up in panels and the like.
+	 * Destroy the content_tree at unmap. Alternatively, we could
+	 * let wlr_scene manage its lifetime automatically, but this
+	 * approach is symmetrical with handle_map() and avoids any
+	 * concern of a dangling pointer in view->content_tree.
 	 */
-out:
-	if (client_request && view->foreign_toplevel) {
-		foreign_toplevel_destroy(view->foreign_toplevel);
-		view->foreign_toplevel = NULL;
+	if (view->content_tree) {
+		wlr_scene_node_destroy(&view->content_tree->node);
+		view->content_tree = NULL;
 	}
 }
 
 static void
-xwayland_view_maximize(struct view *view, bool maximized)
+xwayland_view_maximize(struct view *view, enum view_axis maximized)
 {
 	wlr_xwayland_surface_set_maximized(xwayland_surface_from_view(view),
-		maximized);
+		maximized & VIEW_AXIS_HORIZONTAL, maximized & VIEW_AXIS_VERTICAL);
 }
 
 static void
@@ -834,8 +904,7 @@ xwayland_view_append_children(struct view *self, struct wl_array *children)
 	struct wlr_xwayland_surface *surface = xwayland_surface_from_view(self);
 	struct view *view;
 
-	wl_list_for_each_reverse(view, &self->server->views, link)
-	{
+	wl_list_for_each_reverse(view, &self->server->views, link) {
 		if (view == self) {
 			continue;
 		}
@@ -849,7 +918,7 @@ xwayland_view_append_children(struct view *self, struct wl_array *children)
 		if (!view->surface) {
 			continue;
 		}
-		if (!view->mapped && !view->minimized) {
+		if (!view->mapped) {
 			continue;
 		}
 		if (top_parent_of(view) != surface) {
@@ -860,6 +929,12 @@ xwayland_view_append_children(struct view *self, struct wl_array *children)
 	}
 }
 
+static bool
+xwayland_view_is_modal_dialog(struct view *self)
+{
+	return xwayland_surface_from_view(self)->modal;
+}
+
 static void
 xwayland_view_set_activated(struct view *view, bool activated)
 {
@@ -871,6 +946,14 @@ xwayland_view_set_activated(struct view *view, bool activated)
 	}
 
 	wlr_xwayland_surface_activate(xwayland_surface, activated);
+	/*
+	 * Make sure that the X11-protocol messages (SetInputFocus etc.)
+	 * are sent immediately. This mitigates a race where the XWayland
+	 * server may generate an unwanted FocusOut event for the newly
+	 * activated window, if it receives mouse/pointer events over the
+	 * parallel wayland connection first.
+	 */
+	xwayland_flush(view->server);
 }
 
 static void
@@ -896,17 +979,17 @@ xwayland_view_get_pid(struct view *view)
 static const struct view_impl xwayland_view_impl = {
 	.configure = xwayland_view_configure,
 	.close = xwayland_view_close,
-	.get_string_prop = xwayland_view_get_string_prop,
-	.map = xwayland_view_map,
 	.set_activated = xwayland_view_set_activated,
 	.set_fullscreen = xwayland_view_set_fullscreen,
-	.unmap = xwayland_view_unmap,
 	.maximize = xwayland_view_maximize,
 	.minimize = xwayland_view_minimize,
+	.get_parent = xwayland_view_get_parent,
 	.get_root = xwayland_view_get_root,
 	.append_children = xwayland_view_append_children,
+	.is_modal_dialog = xwayland_view_is_modal_dialog,
 	.get_size_hints = xwayland_view_get_size_hints,
 	.wants_focus = xwayland_view_wants_focus,
+	.offer_focus = xwayland_view_offer_focus,
 	.has_strut_partial = xwayland_view_has_strut_partial,
 	.contains_window_type = xwayland_view_contains_window_type,
 	.get_pid = xwayland_view_get_pid,
@@ -922,9 +1005,7 @@ xwayland_view_create(struct server *server,
 	view->server = server;
 	view->type = LAB_XWAYLAND_VIEW;
 	view->impl = &xwayland_view_impl;
-
-	/* Set to -1 so we always restack the view on map */
-	xwayland_view->stacking_order = -1;
+	view_init(view);
 
 	/*
 	 * Set two-way view <-> xsurface association.  Usually the association
@@ -937,8 +1018,10 @@ xwayland_view_create(struct server *server,
 	xsurface->data = view;
 
 	view->workspace = server->workspaces.current;
-	view->scene_tree = wlr_scene_tree_create(view->workspace->tree);
-	node_descriptor_create(&view->scene_tree->node, LAB_NODE_DESC_VIEW, view);
+	view->scene_tree = wlr_scene_tree_create(
+		view->workspace->view_trees[VIEW_LAYER_NORMAL]);
+	node_descriptor_create(&view->scene_tree->node,
+		LAB_NODE_VIEW, view, /*data*/ NULL);
 
 	CONNECT_SIGNAL(xsurface, view, destroy);
 	CONNECT_SIGNAL(xsurface, view, request_minimize);
@@ -958,16 +1041,17 @@ xwayland_view_create(struct server *server,
 	CONNECT_SIGNAL(xsurface, xwayland_view, set_override_redirect);
 	CONNECT_SIGNAL(xsurface, xwayland_view, set_strut_partial);
 	CONNECT_SIGNAL(xsurface, xwayland_view, set_window_type);
+	CONNECT_SIGNAL(xsurface, xwayland_view, focus_in);
 	CONNECT_SIGNAL(xsurface, xwayland_view, map_request);
 
-	view_init(view);
 	wl_list_insert(&view->server->views, &view->link);
+	view->creation_id = view->server->next_view_creation_id++;
 
 	if (xsurface->surface) {
 		handle_associate(&xwayland_view->associate, NULL);
 	}
 	if (mapped) {
-		xwayland_view_map(view);
+		handle_map(&xwayland_view->base.mappable.map, NULL);
 	}
 }
 
@@ -977,7 +1061,6 @@ handle_new_surface(struct wl_listener *listener, void *data)
 	struct server *server =
 		wl_container_of(listener, server, xwayland_new_surface);
 	struct wlr_xwayland_surface *xsurface = data;
-	wlr_xwayland_surface_ping(xsurface);
 
 	/*
 	 * We do not create 'views' for xwayland override_redirect surfaces,
@@ -990,21 +1073,67 @@ handle_new_surface(struct wl_listener *listener, void *data)
 	}
 }
 
-static void
-sync_atoms(xcb_connection_t *xcb_conn)
+static struct xwayland_view *
+xwayland_view_from_window_id(struct server *server, xcb_window_t id)
 {
+	struct view *view;
+	wl_list_for_each(view, &server->views, link) {
+		if (view->type != LAB_XWAYLAND_VIEW) {
+			continue;
+		}
+		struct xwayland_view *xwayland_view = xwayland_view_from_view(view);
+		if (xwayland_view->xwayland_surface
+				&& xwayland_view->xwayland_surface->window_id == id) {
+			return xwayland_view;
+		}
+	}
+	return NULL;
+}
+
+#define XCB_EVENT_RESPONSE_TYPE_MASK 0x7f
+static bool
+handle_x11_event(struct wlr_xwayland *wlr_xwayland, xcb_generic_event_t *event)
+{
+	switch (event->response_type & XCB_EVENT_RESPONSE_TYPE_MASK) {
+	case XCB_PROPERTY_NOTIFY: {
+		xcb_property_notify_event_t *ev = (void *)event;
+		if (ev->atom == atoms[ATOM_NET_WM_ICON]) {
+			struct server *server = wlr_xwayland->data;
+			struct xwayland_view *xwayland_view =
+				xwayland_view_from_window_id(server, ev->window);
+			if (xwayland_view) {
+				update_icon(xwayland_view);
+			} else {
+				wlr_log(WLR_DEBUG, "icon property changed for unknown window");
+			}
+			return true;
+		}
+		break;
+	}
+	default:
+		break;
+	}
+
+	return false;
+}
+
+static void
+sync_atoms(struct server *server)
+{
+	xcb_connection_t *xcb_conn =
+		wlr_xwayland_get_xwm_connection(server->xwayland);
 	assert(xcb_conn);
 
 	wlr_log(WLR_DEBUG, "Syncing X11 atoms");
-	xcb_intern_atom_cookie_t cookies[WINDOW_TYPE_LEN];
+	xcb_intern_atom_cookie_t cookies[ATOM_COUNT];
 
 	/* First request everything and then loop over the results to reduce latency */
-	for (size_t i = 0; i < WINDOW_TYPE_LEN; i++) {
+	for (size_t i = 0; i < ATOM_COUNT; i++) {
 		cookies[i] = xcb_intern_atom(xcb_conn, 0,
 			strlen(atom_names[i]), atom_names[i]);
 	}
 
-	for (size_t i = 0; i < WINDOW_TYPE_LEN; i++) {
+	for (size_t i = 0; i < ATOM_COUNT; i++) {
 		xcb_generic_error_t *err = NULL;
 		xcb_intern_atom_reply_t *reply =
 			xcb_intern_atom_reply(xcb_conn, cookies[i], &err);
@@ -1014,6 +1143,7 @@ sync_atoms(xcb_connection_t *xcb_conn)
 				atom_names[i], reply->atom);
 		}
 		if (err) {
+			atoms[i] = XCB_ATOM_NONE;
 			wlr_log(WLR_INFO, "Failed to get X11 atom for %s",
 				atom_names[i]);
 		}
@@ -1028,21 +1158,9 @@ handle_server_ready(struct wl_listener *listener, void *data)
 	/* Fire an Xwayland startup script if one (or many) can be found */
 	session_run_script("xinitrc");
 
-	xcb_connection_t *xcb_conn = xcb_connect(NULL, NULL);
-	if (xcb_connection_has_error(xcb_conn)) {
-		wlr_log(WLR_ERROR, "Failed to create xcb connection");
-
-		/* Just clear all existing atoms */
-		for (size_t i = 0; i < WINDOW_TYPE_LEN; i++) {
-			atoms[i] = XCB_ATOM_NONE;
-		}
-		return;
-	}
-
-	wlr_log(WLR_DEBUG, "Connected to xwayland");
-	sync_atoms(xcb_conn);
-	wlr_log(WLR_DEBUG, "Disconnecting from xwayland");
-	xcb_disconnect(xcb_conn);
+	struct server *server =
+		wl_container_of(listener, server, xwayland_server_ready);
+	sync_atoms(server);
 }
 
 static void
@@ -1076,6 +1194,9 @@ xwayland_server_init(struct server *server, struct wlr_compositor *compositor)
 	wl_signal_add(&server->xwayland->events.ready,
 		&server->xwayland_xwm_ready);
 
+	server->xwayland->data = server;
+	server->xwayland->user_event_handler = handle_x11_event;
+
 	if (setenv("DISPLAY", server->xwayland->display_name, true) < 0) {
 		wlr_log_errno(WLR_ERROR, "unable to set DISPLAY for xwayland");
 	} else {
@@ -1095,96 +1216,6 @@ xwayland_server_init(struct server *server, struct wlr_compositor *compositor)
 	}
 }
 
-enum xwayland_view_layer {
-	XWAYLAND_VIEW_HIDDEN,
-	XWAYLAND_VIEW_BOTTOM,
-	XWAYLAND_VIEW_NORMAL,
-	XWAYLAND_VIEW_TOP,
-};
-
-static enum xwayland_view_layer
-get_layer(struct view *view)
-{
-	if (view->workspace != view->server->workspaces.current) {
-		/*
-		 * Until we expose the workspaces to xwayland we need a way to
-		 * ensure that xwayland views on the current workspace are
-		 * always stacked above xwayland views on other workspaces.
-		 *
-		 * If we fail to do so, issues arise in scenarios where we
-		 * change the mouse focus but do not change the (xwayland)
-		 * stacking order.
-		 *
-		 * Reproducer:
-		 * - open at least two xwayland windows which allow scrolling
-		 *   (e.g. urxvt with 'man man')
-		 * - switch to another workspace, open another xwayland window
-		 *   which allows scrolling and maximize it
-		 * - switch back to the previous workspace with the two windows
-		 * - move the mouse to the xwayland window that does *not* have
-		 *   focus
-		 * - start scrolling
-		 * - all scroll events should end up on the maximized window on
-		 *   the other workspace
-		 */
-		return XWAYLAND_VIEW_HIDDEN;
-	} else if (view->shaded) {
-		/*
-		 * Ensure that we don't raise a shaded window to the front
-		 * which then steals mouse events.
-		 */
-		return XWAYLAND_VIEW_HIDDEN;
-	} else if (view_is_always_on_bottom(view)) {
-		return XWAYLAND_VIEW_BOTTOM;
-	} else if (view_is_always_on_top(view)) {
-		return XWAYLAND_VIEW_TOP;
-	} else {
-		return XWAYLAND_VIEW_NORMAL;
-	}
-}
-
-void
-xwayland_adjust_stacking_order(struct server *server)
-{
-	if (!server->xwayland) {
-		/* This happens when windows are unmapped on exit */
-		return;
-	}
-
-	int stacking_order = 0;
-	bool update = false;
-
-	/*
-	 * Iterate over the windows from bottom to top and notify their
-	 * stacking order to xwayland if we detect updates in it. Note
-	 * that server->views are sorted from top to bottom but doesn't
-	 * consider always-on-{top,bottom} windows.
-	 */
-	for (enum xwayland_view_layer layer = XWAYLAND_VIEW_HIDDEN;
-			layer <= XWAYLAND_VIEW_TOP; layer++) {
-		struct view *view;
-		wl_list_for_each_reverse(view, &server->views, link) {
-			if (view->type != LAB_XWAYLAND_VIEW
-					|| layer != get_layer(view)) {
-				continue;
-			}
-
-			struct xwayland_view *xwayland_view =
-				xwayland_view_from_view(view);
-
-			/* On detecting update, restack all the views above it */
-			update |= xwayland_view->stacking_order != stacking_order;
-			if (update) {
-				wlr_xwayland_surface_restack(
-					xwayland_view->xwayland_surface,
-					NULL, XCB_STACK_MODE_ABOVE);
-			}
-			xwayland_view->stacking_order = stacking_order;
-			stacking_order++;
-		}
-	}
-}
-
 void
 xwayland_reset_cursor(struct server *server)
 {
@@ -1417,3 +1448,13 @@ xwayland_update_workarea(struct server *server)
 	};
 	wlr_xwayland_set_workareas(server->xwayland, &workarea, 1);
 }
+
+void
+xwayland_flush(struct server *server)
+{
+	if (!server->xwayland || !server->xwayland->xwm) {
+		return;
+	}
+
+	xcb_flush(wlr_xwayland_get_xwm_connection(server->xwayland));
+}
